SuperStrict

Framework wx.wxApp
Import wx.wxFrame
Import wx.wxScrolledWindow
Import wx.wxStaticBitmap
Import wx.wxListCtrl
Import wx.wxGauge
Import wx.wxStaticText
Import wx.wxButton
Import wx.wxDialog
Import wx.wxStaticBox
Import wx.wxCheckBox
Import wx.wxBitmapButton
Import wx.wxRadioButton
Import wx.wxColourDialog
Import wx.wxSpinCtrl
Import wx.wxFileDialog
Import wx.wxMessageDialog
Import wx.wxAboutBox
Import wx.wxTimer
Import wx.wx

Import BRL.StandardIO
Import BRL.Random
Import BRL.Retro
Import BRL.FileSystem
Import BRL.PNGLoader

Import chaos.mathext

Import "Fractal Core.bmx"

Const wxID_EDIT_FRACTAL:Int	= 100
Const wxID_EDIT_SCENE:Int	= 101
Const wxID_EDIT_RENDERER:Int	= 102
Const wxID_RENDER:Int		= 103

Const wxID_SCROLLWINDOW:Int	= 200

Global MainApp:TApp = New TApp

MainApp.Run()
 
Type TApp Extends wxApp
	Field MainWin:TMainWindow
	
	Method OnInit:Int()
		wxImage.AddHandler( New wxPNGHandler )
		wxImage.AddHandler( New wxJPEGHandler )
		wxImage.AddHandler( New wxICOHandler )
		
		MainWin = TMainWindow( New TMainWindow.Create( ,, "Chaos Fractal",,, 800, 600, wxDEFAULT_FRAME_STYLE | wxMAXIMIZE ) )
		MainWin.Show( True )
		MainWin.InitListControl()
		
		Return True
	End Method
End Type

Type TMainWindow Extends wxFrame
	Const LISTCTRL_W:Int = 200
	
	Field ProgressWindow:TProgressWindow
	
	Field ProjectListControl:TProjectListControl
	Field CurrentProject:TProject
	
	Field MenuBar:TCustomMenuBar
	
	Field ScrolledWindow:TScrolledImageWindow
	
	Method OnInit()
		'------------ Menu
		InitMenuBar()
		
		'------------ Statusbar
		InitStatusBar()
		
		'------------ Work space
		InitMainWindow()
		
		
		SetCurrentProject( Null )
		
		Centre()
	End Method
	
	Method InitMenuBar()
		MenuBar = TCustomMenuBar( New TCustomMenuBar.Create() )
		
		MenuBar.OnInit()
		
		Connect( wxID_NEW,		wxEVT_COMMAND_MENU_SELECTED, DispatchNew )
		Connect( wxID_OPEN,		wxEVT_COMMAND_MENU_SELECTED, DispatchOpen )
		Connect( wxID_SAVE,		wxEVT_COMMAND_MENU_SELECTED, DispatchSave )
		Connect( wxID_SAVEAS,		wxEVT_COMMAND_MENU_SELECTED, DispatchSaveAs )
		Connect( wxID_CLOSE,		wxEVT_COMMAND_MENU_SELECTED, DispatchClose )
		Connect( wxID_RENDER,		wxEVT_COMMAND_MENU_SELECTED, DispatchRender )
		Connect( wxID_EDIT_RENDERER,	wxEVT_COMMAND_MENU_SELECTED, DispatchEditRender )
		Connect( wxID_EDIT_FRACTAL,	wxEVT_COMMAND_MENU_SELECTED, DispatchEditFractal )
		Connect( wxID_EDIT_SCENE,	wxEVT_COMMAND_MENU_SELECTED, DispatchEditScene )
		Connect( wxID_HELP,		wxEVT_COMMAND_MENU_SELECTED, DispatchHelp )
		Connect( wxID_ABOUT,		wxEVT_COMMAND_MENU_SELECTED, DispatchAbout )
		
		SetMenuBar( MenuBar )
	End Method
	
	Method InitStatusBar()
		Local Statusbar:wxStatusBar = New wxStatusBar.Create( Self, 100 )
		
		SetStatusBar( Statusbar )
	End Method
	
	Method InitMainWindow()
		'------------ Main panel
		Local Panel:wxPanel = New wxPanel.Create( Self, -1 )
		
		'------------ Scrollable image window
		ScrolledWindow:TScrolledImageWindow = TScrolledImageWindow( New TScrolledImageWindow.Create( Panel, wxID_SCROLLWINDOW,,,,, wxVSCROLL | wxHSCROLL | wxBORDER_SUNKEN | wxFULL_REPAINT_ON_RESIZE ) )
		
		'------------ Listcontrol
		ProjectListControl = TProjectListControl( New TProjectListControl.Create( Panel, -1,,, LISTCTRL_W,, wxLC_REPORT | wxLC_NO_HEADER | wxLC_EDIT_LABELS | wxLC_SINGLE_SEL ) )
		
		'------------ Fill and set sizer
		Local Sizer:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		
		Sizer.AddSpacer( 3 )
		Sizer.Add( ProjectListControl, 0, wxEXPAND | wxRIGHT, 5 )
		Sizer.Add( ScrolledWindow, 1, wxEXPAND )
		
		Panel.SetSizer( Sizer )
		
		ConnectAny( wxEVT_CLOSE, DispatchClose )
		
		Local IconBundle:wxIconBundle = New wxIconBundle.Create( "Media/Chaos Fractal.ico", wxBITMAP_TYPE_ICO )
		
		SetIcons( IconBundle )
	End Method
	
	Method InitListControl()
		Local FolderWin:TLoadFolderWindow = TLoadFolderWindow( New TLoadFolderWindow.Create( Self,, "Loading",,, 250, 60, wxCAPTION | wxFRAME_TOOL_WINDOW ) )
		FolderWin.Show( True )
		
		ReadFractalFolder()
		
		FolderWin.DeInit()
		
		SetCurrentProject( TProject( ProjectListControl.GetItemData( 0 ) ) )
	End Method
	
	Method ReadFractalFolder()
		Local Folder:Int = ReadDir( "Fractals" )
		
		If Folder Then
			Local CurrFile:String = NextFile( Folder )
			While CurrFile <> ""
				If FileType( "Fractals/" + CurrFile ) = FILETYPE_FILE Then
					If ExtractExt( CurrFile ).ToLower() = "fractal" Then
						OpenFractal( "Fractals/" + CurrFile, False )
						MainApp.Yield()
					EndIf
				EndIf
				
				CurrFile = NextFile( Folder )
			Wend
			
			CloseDir( Folder )
		EndIf
	End Method
	
	Method SetCurrentProject( Project:TProject, Mark:Int = True )
		CurrentProject = Project
		
		If Mark Then ProjectListControl.MarkProject( Project )
		MenuBar.EnableMenus( Project <> Null )
		ScrolledWindow.SetProject( Project )
	End Method
	
	Method CreateRenderDialog()
		Local RenderDialog:TRenderDialog = TRenderDialog( New TRenderDialog.Create_( Self, -1, "Render settings",,, 320, 147 ) )
		RenderDialog.Show( True )
	End Method
	
	Method CreateFractalDialog()
		Local FractalDialog:TFractalDialog = TFractalDialog( New TFractalDialog.Create_( Self, -1, "Fractal settings",,, 300, 155 ) )
		FractalDialog.Show( True )
	End Method
	
	Method CreateSceneDialog()
		Local SceneDialog:TSceneDialog = TSceneDialog( New TSceneDialog.Create_( Self, -1, "Scene settings",,, 300, 185 ) )
		SceneDialog.Show( True )
	End Method
	
	Method CreateProgressWindow()
		ProgressWindow = TProgressWindow( New TProgressWindow.Create( Self,, "Progress",,, 400, 50, wxCAPTION | wxFRAME_TOOL_WINDOW | wxFRAME_FLOAT_ON_PARENT ) )
		ProgressWindow.Show( True )
	End Method
	
	Method FreeProgressWindow()
		ProgressWindow.Destroy()
		ProgressWindow = Null
	End Method
	
	Method RenderFractal()
		CreateProgressWindow()
		CurrentProject.Render()
		FreeProgressWindow()
		
		ScrolledWindow.SetProject( CurrentProject )
	End Method
	
	Method OpenFractal( Path:String, SetCurrent:Int = True )
		Local Name:String = StripAll( Path )
		
		Local Project:TProject = TProject.Create( Name )
		Project.Path = Path
		
		Project.SetStandardValues()
		Project.LoadFromFile()
		
		ProjectListControl.AddProject( Project )
		
		If SetCurrent Then SetCurrentProject( Project )
	End Method
	
	Function DispatchNew( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		Local Project:TProject = TProject.Create( "Untitled" )
		Project.SetStandardValues()
		Project.RandomizeFractal()
		Project.Changed = True
		
		Local ListItem:Int = MainWindow.ProjectListControl.AddProject( Project )
		
		MainWindow.ProjectListControl.EditLabel( ListItem )
		MainWindow.SetCurrentProject( Project )
	End Function
	
	Function DispatchOpen( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		Local Path:String = wxFileSelector( "Open", "Fractals",,, "Fractal files (*.fractal)|*.fractal", wxFD_OPEN | wxFD_FILE_MUST_EXIST, MainWindow )
		
		If Path <> "" Then MainWindow.OpenFractal( Path )
	End Function
	
	Function DispatchSave( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		MainWindow.CurrentProject.SaveToFile()
		MainWindow.CurrentProject.MarkUnchanged()
	End Function
	
	Function DispatchSaveAs( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		Local Path:String = wxFileSelector( "Save as", "Fractals", MainWindow.CurrentProject.Name,, "Fractal files (*.fractal)|*.fractal", wxFD_SAVE | wxFD_OVERWRITE_PROMPT, MainWindow )
		
		If Path <> "" Then
			MainWindow.CurrentProject.SetPath( Path )
			MainWindow.CurrentProject.SaveToFile()
			
			MainWindow.CurrentProject.MarkUnchanged()
		EndIf
	End Function
	
	Function DispatchClose( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		For Local I:Int = 0 To MainWindow.ProjectListControl.GetItemCount() - 1
			Local Project:TProject = TProject( MainWindow.ProjectListControl.GetItemData( I ) )
			
			If Project.Changed Then
				Local Dialog:wxMessageDialog = New wxMessageDialog.Create( MainWindow, Project.Name + " has been changed. Would you like to save changes now?", "Chaos Fractal", wxYES_NO | wxCANCEL | wxICON_QUESTION )
				
				Local Result:Int = Dialog.ShowModal()
				
				If Result = wxID_YES Then
					Project.SaveToFile()
				ElseIf Result = wxID_CANCEL Then
					Local CloseEvent:wxCloseEvent = wxCloseEvent( Event )
					
					If CloseEvent <> Null And CloseEvent.CanVeto() Then CloseEvent.Veto()
					
					Return
				EndIf
			EndIf
		Next
		
		End
	End Function
	
	Function DispatchEditFractal( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		MainWindow.CreateFractalDialog()
	End Function
	
	Function DispatchEditScene( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		MainWindow.CreateSceneDialog()
	End Function
	
	Function DispatchEditRender( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		MainWindow.CreateRenderDialog()
	End Function
	
	Function DispatchRender( Event:wxEvent )
		Local MainWindow:TMainWindow = TMainWindow( Event.Parent )
		
		MainWindow.RenderFractal()
	End Function
	
	Function DispatchHelp( Event:wxEvent )
		OpenURL( RealPath( "ReadMe.txt" ) )
	End Function
	
	Function DispatchAbout( Event:wxEvent )
		Local Info:wxAboutDialogInfo = New wxAboutDialogInfo.Create()
		
		Info.SetName( "Chaos Fractal" )
		Info.SetDescription( "Free fractals for everyone!~n~nVisit www.blitzforum.de for other cool stuff!" )
		Info.SetVersion( "1.00 Beta" )
		
		wxAboutBox( Info )
	End Function
End Type

Type TScrolledImageWindow Extends wxScrolledWindow
	Field Image:wxBitmap
	
	Method OnInit()
		Connect( wxID_SCROLLWINDOW, wxEVT_PAINT, OnPaint )
	End Method
	
	Method SetProject( Project:TProject )
		If Project Then
			Local RenderedImage:TPixmap = Project.GetImage()
			Image = Project.GetBitmap()
			
			Local BGColor:TVector = Project.Camera.BackgroundColor
			Local Width:Int, Height:Int
			
			GetClientSize( Width, Height )
			
			If Image Then
				SetScrollbars( 1, 1, RenderedImage.Width, RenderedImage.Height )
			Else
				SetScrollbars( 1, 1, 1, 1 )
			EndIf
			
			Scroll( 1, 1 )
			
			SetBackgroundColour( wxColour.CreateColour( BGColor.X, BGColor.Y, BGColor.Z ) )
		Else
			Image = Null
			SetBackgroundColour( New wxColour.CreateNamedColour( "#EDEDED" ) )
			SetScrollbars( 1, 1, 1, 1 )
		EndIf
		
		Update()
	End Method
	
	Function OnPaint( Event:wxEvent )
		Local ScrolledWindow:TScrolledImageWindow = TScrolledImageWindow( Event.Parent )
		
		Local DC:wxPaintDC = New wxPaintDC.Create( ScrolledWindow )
		Assert DC <> Null, "Could not create device context!"
		
		If ScrolledWindow.Image Then
			ScrolledWindow.DoPrepareDC( DC )
			
			Local DrawX:Int, DrawY:Int, Width:Int, Height:Int
					
			ScrolledWindow.GetClientSize( Width, Height )
			
			If ScrolledWindow.Image.GetWidth()  < Width  Then DrawX = ( Width  - ScrolledWindow.Image.GetWidth()  )/2
			If ScrolledWindow.Image.GetHeight() < Height Then DrawY = ( Height - ScrolledWindow.Image.GetHeight() )/2
			
			DC.DrawBitmap( ScrolledWindow.Image, DrawX, DrawY, False )
		EndIf
		
		DC.Free()
	End Function
End Type

Type TCustomMenuBar Extends wxMenuBar
	Field FileMenu:wxMenu
	Field EditMenu:wxMenu
	Field HelpMenu:wxMenu
	
	Const FILE:Int	= 0
	Const EDIT:Int	= 1
	Const HELP:Int	= 2
	
	Const CAP_NEW:Int		= 0
	Const CAP_OPEN:Int	= 1
	Const CAP_SAVE:Int	= 2
	Const CAP_SAVEAS:Int	= 3
	Const CAP_CLOSE:Int	= 4
	
	Const CAP_EDIT_FRACTAL:Int	= 0
	Const CAP_EDIT_SCENE:Int	= 1
	Const CAP_EDIT_RENDERER:Int	= 2
	Const CAP_RENDER:Int		= 3
	
	Const CAP_HELP:Int	= 0
	Const CAP_ABOUT:Int	= 1
	
	Global BCAPTION:String[] = [ "&File", "&Edit", "&Help" ]
	Global FCAPTION:String[] = [ "&New~tCtrl-N", "&Open~tCtrl-O", "&Save~tCtrl-S", "Save &as...", "&Exit~tCtrl-Q" ]
	Global ECAPTION:String[] = [ "Edit &fractal", "Edit &scene", "R&ender settings", "&Render!~tCtrl-R" ]
	Global HCAPTION:String[] = [ "&Help~tF1", "&About" ]
	
	Method OnInit()
		FileMenu:wxMenu = New wxMenu.Create()
			FileMenu.Append( wxID_NEW,    FCAPTION[ CAP_NEW ] )
			FileMenu.Append( wxID_OPEN,   FCAPTION[ CAP_OPEN ] )
			FileMenu.Append( wxID_SAVE,   FCAPTION[ CAP_SAVE ] )
			FileMenu.Append( wxID_SAVEAS, FCAPTION[ CAP_SAVEAS ] )
			FileMenu.AppendSeparator()
			FileMenu.Append( wxID_CLOSE,  FCAPTION[ CAP_CLOSE ] )
		
		EditMenu = New wxMenu.Create()
			EditMenu.Append( wxID_EDIT_FRACTAL,  ECAPTION[ CAP_EDIT_FRACTAL ] )
			EditMenu.Append( wxID_EDIT_SCENE,    ECAPTION[ CAP_EDIT_SCENE ] )
			EditMenu.Append( wxID_EDIT_RENDERER, ECAPTION[ CAP_EDIT_RENDERER ] )
			EditMenu.AppendSeparator()
			EditMenu.Append( wxID_RENDER,        ECAPTION[ CAP_RENDER ] )
		
		HelpMenu:wxMenu = New wxMenu.Create()
			HelpMenu.Append( wxID_HELP,  HCAPTION[ CAP_HELP ] )
			HelpMenu.AppendSeparator()
			HelpMenu.Append( wxID_ABOUT, HCAPTION[ CAP_ABOUT ] )
		
		Append( FileMenu, BCAPTION[ FILE ] )
		Append( EditMenu, BCAPTION[ EDIT ] )
		Append( HelpMenu, BCAPTION[ HELP ] )
	End Method
	
	Method EnableMenus( Mode:Int )
		EnableTop( FindMenu( BCAPTION[ EDIT ] ), Mode )
		
		FileMenu.Enable( FileMenu.FindItem( FCAPTION[ CAP_SAVE ] ),   Mode )
		FileMenu.Enable( FileMenu.FindItem( FCAPTION[ CAP_SAVEAS ] ), Mode )
	End Method
End Type

Type TProjectListControl Extends wxListCtrl
	Field Width:Int
	Field NewSelection:Int
	
	Method OnInit()
		Local IconList:wxImageList = New wxImageList.Create( 16, 16, True )
		
		IconList.Add( wxBitmap.CreateFromFile( "Media/Image.png",     wxBITMAP_TYPE_PNG ) )
		IconList.Add( wxBitmap.CreateFromFile( "Media/Animation.png", wxBITMAP_TYPE_PNG ) )
		
		AssignImageList( IconList, wxIMAGE_LIST_SMALL )
		
		Local Dummy:Int
		GetClientSize( Width, Dummy )
		
		InsertColumn( 0, "",, Width - 10 ) 'Insert an invisible column to make the items visible
		
		ConnectAny( wxEVT_COMMAND_LIST_END_LABEL_EDIT,	DispatchLabelEdit )
		ConnectAny( wxEVT_COMMAND_LIST_ITEM_SELECTED,		DispatchSelection )
		ConnectAny( wxEVT_COMMAND_LIST_ITEM_DESELECTED,	DispatchDeselection )
	End Method
	
	Method AddProject:Int( Project:TProject )
		Local Item:wxListItem = New wxListItem.Create()
		
		Item.SetText( Project.Name )
		Item.SetImage( 0 )
		Item.SetData( Project )
		
		Return InsertItem( Item )
	End Method
	
	Method MarkProject( Project:TProject )
		If Project Then
			Local Index:Int = FindItem( -1, Project.Name )
			
			For Local I:Int = 0 To GetItemCount()
				If I = Index Then
					SetItemState( I, wxLIST_STATE_SELECTED | wxLIST_STATE_FOCUSED, wxLIST_STATE_SELECTED | wxLIST_STATE_FOCUSED )
				Else
					SetItemState( I, 0, wxLIST_STATE_SELECTED | wxLIST_STATE_FOCUSED )
				EndIf
			Next
		EndIf
	End Method
	
	Function DispatchLabelEdit( Event:wxEvent )
		Local ListEvent:wxListEvent = wxListEvent( Event )
		
		Local ListCtrl:TProjectListControl = TProjectListControl( ListEvent.Parent )
		
		Local Item:wxListItem	= ListEvent.GetItem()
		Local Name:String		= ListEvent.GetLabel()
		Local Index:Int		= ListEvent.GetIndex()
		
		Local Project:TProject = TProject( Item.GetData() )
		
		If ListEvent.IsEditCancelled() Then Name = Project.GetName()
		
		Local FirstFind:Int = ListCtrl.FindItem( -1, Name )
		Local NextFind:Int  = ListCtrl.FindItem( Index + 1, Name )
		
		'Condition, to make sure, that a project isn't renamed into an existing project
		'Looks a bit strange, because wxWidgets hast a wierd opinion on when an edit has been cancelled
		If ( FirstFind <> -1 ) ~ ( FirstFind = Index ) Or NextFind <> -1 Then
			ListEvent.Veto()
			
			ListCtrl.EditLabel( ListEvent.GetIndex() )
		Else
			Project.SetName( Name )
			Project.SetPath( "Fractals/" + Name + ".fractal" )
		EndIf
	End Function
	
	Function DispatchSelection( Event:wxEvent )
		Local ListEvent:wxListEvent = wxListEvent( Event )
		
		Local MainWindow:TMainWindow = TMainWindow( wxWindow( ListEvent.Parent ).GetParent().GetParent() )
		Local Project:TProject = TProject( ListEvent.GetItem().GetData() )
		
		MainWindow.SetCurrentProject( Project, False )
		MainWindow.ProjectListControl.NewSelection = True
	End Function
	
	Function DispatchDeselection( Event:wxEvent )
		Local ListEvent:wxListEvent = wxListEvent( Event )

		Local ListControl:TProjectListControl = TProjectListControl( Event.Parent )
		If ListControl.GetSelectedItemCount() = 0 Then

			Local MainWindow:TMainWindow = TMainWindow( wxWindow( ListEvent.Parent ).GetParent().GetParent() )

			MainWindow.ProjectListControl.NewSelection = False
			TMainWindow( ListControl.GetParent().GetParent() ).SetCurrentProject( Null )
		EndIf
	End Function
End Type

Type TLoadFolderWindow Extends wxFrame
	Const LABEL_CAPTION:String = "Loading fractal files, please wait"
	
	Field Label:wxStaticText
	Field Timer:wxTimer
	Field DotCount:Int = 1
	
	Method OnInit()
		Local Panel:wxPanel = New wxPanel.Create( Self )
		
		Local HSizer:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		
		Label = New wxStaticText.Create( Panel, -1, LABEL_CAPTION )
		
		HSizer.Add( Label, 0, wxALIGN_CENTER | wxLEFT, 30 )
		
		Panel.SetSizer( HSizer )
		
		Timer = New wxTimer.Create( Self )
		Timer.Start( 400 )
		
		ConnectAny( wxEVT_TIMER, DispatchTimerEvent )
		
		Center()
	End Method
	
	Method DeInit()
		Timer.Stop()
		
		Destroy()
	End Method
	
	Function DispatchTimerEvent( Event:wxEvent )
		Local FolderWin:TLoadFolderWindow = TLoadFolderWindow( Event.Parent )
		
		FolderWin.DotCount = ( FolderWin.DotCount Mod 3 ) + 1
		
		FolderWin.Label.SetLabel( LABEL_CAPTION + "."[ .. FolderWin.DotCount ].Replace( " ", "." ) )
	End Function
End Type


Type TProgressWindow Extends wxFrame
	Field Gauge:wxGauge
	Field ContinueRender:Int
	
	Method OnInit()
		'----------- Set the continue flag
		ContinueRender = True
		
		'----------- Main panel
		Local Panel:wxPanel = New wxPanel.Create( Self )
		
		'----------- Sizer
		Local HSizer:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		Local VSizer:wxBoxSizer = New wxBoxSizer.Create( wxVERTICAL )
		
		'----------- Progress bar
		Gauge = New wxGauge.Create( Panel, -1, 100 )
		Gauge.SetValue( 50 )
		
		'----------- Cancel button
		Local Button:wxButton = New wxButton.Create( Panel, wxID_ABORT, "Cancel" )
		
		'----------- Set eventhandler that dispatches the abort event
		Connect( wxID_ABORT, wxEVT_COMMAND_BUTTON_CLICKED, OnAbort )
		
		'----------- Prepare sizer and set it
		VSizer.Add( Gauge, 1, wxEXPAND )
		
		HSizer.AddSizer( VSizer, 1, wxEXPAND | wxRIGHT, 3 )
		HSizer.Add( Button, 0, wxEXPAND | wxRIGHT | wxTOP | wxBOTTOM, 2 )
		
		Panel.SetSizer( HSizer )
		
		Center()
	End Method
	
	Function OnAbort( Event:wxEvent )
		Local ProgressWindow:TProgressWindow = TProgressWindow( Event.Parent )
		
		ProgressWindow.ContinueRender = False
	End Function
End Type

Type TFractalDialog Extends wxDialog
	Const wxID_VTEXTCTRL_C:Int	= 300
	Const wxID_BUTTON_COLOR:Int	= 301
	Const wxID_RADIO_SQUARED:Int	= 302
	Const wxID_RADIO_CUBIC:Int	= 303
	
	Field ParentProject:TProject
	
	Field C:TVector
	Field FractalColor:TVector
	Field FractalType:Int
	
	Field Changed:Int
	
	Method OnInit()
		ParentProject = TMainWindow( GetParent() ).CurrentProject
		
		Local VSizer:wxBoxSizer = New wxBoxSizer.Create( wxVERTICAL )
		
		Local HSizer1:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		Local HSizer2:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		
		Local GSizer:wxFlexGridSizer = New wxFlexGridSizer.CreateRC( 4, 2, 10, 15 )
		
		'-------------- Fractal constant
		Local Label1:wxStaticText = New wxStaticText.Create( Self, -1, "C  = " )
		Local VectorInput:TVectorTextCtrl = TVectorTextCtrl( New TVectorTextCtrl.Create( Self, wxID_VTEXTCTRL_C,,,,, 21 ) )
		
		C = ParentProject.Fractal.C
		
		VectorInput.SetScalarCount( 4 )
		VectorInput.SetScalars( [ C.X, C.Y, C.Z, C.W ], 3 )
		
		GSizer.Add( Label1, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.Add( VectorInput, 0, wxALIGN_CENTER_VERTICAL | wxEXPAND )
		
		'--------------- Color
		FractalColor = ParentProject.Fractal.DiffuseColor
		
		Local ColorImage:wxImage = wxImage.CreateEmptyImage( 60, 12 )
		ColorImage.SetRGBRange( 0, 0, 60, 12, FractalColor.X, FractalColor.Y, FractalColor.Z )
		
		Local ColorBitmap:wxBitmap = wxBitmap.CreateFromImage( ColorImage )
		
		Local Label2:wxStaticText = New wxStaticText.Create( Self, -1, "Color:" )
		Local ImgButton:wxBitmapButton = New wxBitmapButton.Create( Self, wxID_BUTTON_COLOR, ColorBitmap,,, 70, 20 )
		
		GSizer.Add( Label2, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.Add( ImgButton, 0, wxALIGN_CENTER_VERTICAL )
		
		'--------------- Fractal type
		FractalType = ParentProject.Fractal.FractalType
		
		Local Label3:wxStaticText = New wxStaticText.Create( Self, -1, "Fraktal type:" )
		
		Local Radio1:wxRadioButton = New wxRadioButton.Create( Self, wxID_RADIO_SQUARED, "Squared",,,,, wxRB_GROUP )
		Local Radio2:wxRadioButton = New wxRadioButton.Create( Self, wxID_RADIO_CUBIC,   "Cubic" )
		
		If FractalType = FRACTAL_SQUARED Then
			Radio1.SetValue( True )
		Else
			Radio2.SetValue( True )
		EndIf
		
		HSizer1.Add( Radio1, 0, wxALIGN_CENTER_VERTICAL | wxRIGHT, 10 )
		HSizer1.Add( Radio2, 0, wxALIGN_CENTER_VERTICAL | wxRIGHT, 10 )
		
		GSizer.Add( Label3, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.AddSizer( HSizer1, 0, wxALIGN_CENTER_VERTICAL )
		
		'--------------- Buttons
		Local Button1:wxButton = New wxButton.Create( Self, wxID_OK, "OK" )
		Local Button2:wxButton = New wxButton.Create( Self, wxID_CANCEL, "Cancel" )
		
		HSizer2.Add( Button1, 0, wxRIGHT, 10 )
		HSizer2.Add( Button2, 0, wxLEFT, 10 )
		
		'--------------- Vertical sizer
		VSizer.AddSpacer( 15 )
		VSizer.AddSizer( GSizer, 1, wxALIGN_CENTER_HORIZONTAL | wxEXPAND | wxLEFT, 30 )
		VSizer.AddSizer( HSizer2, 0, wxALIGN_CENTER | wxBOTTOM, 5 )
		
		Connect( wxID_BUTTON_COLOR, wxEVT_COMMAND_BUTTON_CLICKED, DispatchColorPick )
		Connect( wxID_VTEXTCTRL_C, wxEVT_COMMAND_TEXT_UPDATED, DispatchVectorUpdate )
		Connect( wxID_RADIO_SQUARED, wxEVT_COMMAND_RADIOBUTTON_SELECTED, DispatchSquaredSelection )
		Connect( wxID_RADIO_CUBIC, wxEVT_COMMAND_RADIOBUTTON_SELECTED, DispatchCubicSelection )
		
		SetSizer( VSizer )
		
		Center()
		
		If ShowModal() = wxID_OK And Changed Then
			ParentProject.Fractal.C = C
			ParentProject.Fractal.DiffuseColor = FractalColor
			ParentProject.Fractal.FractalType  = FractalType
			ParentProject.MarkChanged()
			
			Destroy()
			
			TMainWindow( GetParent() ).RenderFractal()
			
			Return
		EndIf
		
		Destroy()
	End Method
	
	Function DispatchColorPick( Event:wxEvent )
		Local ImgButton:wxBitmapButton = wxBitmapButton( Event.GetEventObject() )
		
		Local ColorPicker:wxColourDialog = New wxColourDialog.Create( wxWindow( Event.Parent ) )
		
		If ColorPicker.ShowModal() = wxID_OK Then
			Local Color:wxColour = ColorPicker.GetColourData().GetColour()
			
			Local ColorImage:wxImage = wxImage.CreateEmptyImage( 60, 12 )
			ColorImage.SetRGBRange( 0, 0, 60, 12, Color.Red(), Color.Green(), Color.Blue() )
			
			ImgButton.SetBitmapLabel( New wxBitmap.CreateFromImage( ColorImage ) )
			
			TFractalDialog( Event.Parent ).FractalColor = Vec3( Color.Red(), Color.Green(), Color.Blue() )
			TFractalDialog( Event.Parent ).Changed = True
		EndIf
	End Function
	
	Function DispatchVectorUpdate( Event:wxEvent )
		Local VectorCtrl:TVectorTextCtrl = TVectorTextCtrl( Event.GetEventObject() )
		Local Values:Float[] = VectorCtrl.GetScalars()
		
		TFractalDialog( Event.Parent ).C = Vec4( Values[ 0 ], Values[ 1 ], Values[ 2 ], Values[ 3 ] )
		TFractalDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchSquaredSelection( Event:wxEvent )
		TFractalDialog( Event.Parent ).FractalType = FRACTAL_SQUARED
		TFractalDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchCubicSelection( Event:wxEvent )
		TFractalDialog( Event.Parent ).FractalType = FRACTAL_CUBIC
		TFractalDialog( Event.Parent ).Changed = True
	End Function
End Type

Type TSceneDialog Extends wxDialog
	Const wxID_VTEXTCTRL_FROTATION:Int	= 300
	Const wxID_VTEXTCTRL_LPOSITION:Int	= 301
	Const wxID_VTEXTCTRL_CPOSITION:Int	= 302
	Const wxID_BUTTON_COLOR:Int		= 303
	
	Field ParentProject:TProject
	
	Field FractalRotation:TVector
	Field LightPosition:TVector
	Field CameraPosition:TVector
	Field BackgroundColor:TVector
	
	Field Changed:Int
	
	Method OnInit()
		ParentProject = TMainWindow( GetParent() ).CurrentProject
		
		Local GSizer:wxFlexGridSizer = New wxFlexGridSizer.CreateRC( 4, 2, 7, 5 )
		Local VSizer:wxBoxSizer = New wxBoxSizer.Create( wxVERTICAL )
		Local HSizer:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		
		GSizer.AddGrowableCol( 1 )
		
		'-------------- Fractal rotation
		FractalRotation = ParentProject.Fractal.GetRotation()
		
		Local Label1:wxStaticText = New wxStaticText.Create( Self, -1, "Fraktal rotation:" )
		Local VectorCtrl1:TVectorTextCtrl = TVectorTextCtrl( New TVectorTextCtrl.Create( Self, wxID_VTEXTCTRL_FROTATION ) )
		
		VectorCtrl1.SetScalarCount( 3 )
		VectorCtrl1.SetScalars( [ FractalRotation.X, FractalRotation.Y, FractalRotation.Z ], 2 )
		
		GSizer.Add( Label1, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL )
		GSizer.Add( VectorCtrl1, 0, wxALIGN_LEFT | wxALIGN_CENTER_VERTICAL | wxEXPAND )
		
		'-------------- Light position
		LightPosition = ParentProject.Camera.LightPosition
		
		Local Label2:wxStaticText = New wxStaticText.Create( Self, -1, "Light position:" )
		Local VectorCtrl2:TVectorTextCtrl = TVectorTextCtrl( New TVectorTextCtrl.Create( Self, wxID_VTEXTCTRL_LPOSITION ) )
		
		VectorCtrl2.SetScalarCount( 3 )
		VectorCtrl2.SetScalars( [ LightPosition.X, LightPosition.Y, LightPosition.Z ], 2 )
		
		GSizer.Add( Label2, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL )
		GSizer.Add( VectorCtrl2, 0, wxALIGN_LEFT | wxALIGN_CENTER_VERTICAL | wxEXPAND )
		
		'-------------- Camera position
		CameraPosition = ParentProject.Camera.Position
		
		Local Label3:wxStaticText = New wxStaticText.Create( Self, -1, "Camera position:" )
		Local VectorCtrl3:TVectorTextCtrl = TVectorTextCtrl( New TVectorTextCtrl.Create( Self, wxID_VTEXTCTRL_CPOSITION ) )
		
		VectorCtrl3.SetScalarCount( 3 )
		VectorCtrl3.SetScalars( [ CameraPosition.X, CameraPosition.Y, CameraPosition.Z ], 2 )
		
		GSizer.Add( Label3, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL )
		GSizer.Add( VectorCtrl3, 0, wxALIGN_LEFT | wxALIGN_CENTER_VERTICAL | wxEXPAND )
		
		'-------------- Backgroundcolor
		BackgroundColor = ParentProject.Camera.BackgroundColor
		
		Local ColorImage:wxImage = wxImage.CreateEmptyImage( 60, 12 )
		ColorImage.SetRGBRange( 0, 0, 60, 12, BackgroundColor.X, BackgroundColor.Y, BackgroundColor.Z )
		
		Local ColorBitmap:wxBitmap = wxBitmap.CreateFromImage( ColorImage )
		
		Local Label4:wxStaticText = New wxStaticText.Create( Self, -1, "Backgroundcolor:" )
		Local ImgButton:wxBitmapButton = New wxBitmapButton.Create( Self, wxID_BUTTON_COLOR, ColorBitmap,,, 70, 20 )
		
		GSizer.Add( Label4, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL )
		GSizer.Add( ImgButton, 0, wxALIGN_LEFT | wxALIGN_CENTER_VERTICAL )
		
		'--------------- Buttons
		Local Button1:wxButton = New wxButton.Create( Self, wxID_OK, "OK" )
		Local Button2:wxButton = New wxButton.Create( Self, wxID_CANCEL, "Cancel" )
		
		HSizer.Add( Button1, 0, wxRIGHT, 10 )
		HSizer.Add( Button2, 0, wxLEFT, 10 )
		
		'--------------- Vertical sizer
		VSizer.AddSpacer( 15 )
		VSizer.AddSizer( GSizer, 1, wxALIGN_CENTER_HORIZONTAL | wxEXPAND | wxLEFT | wxRIGHT, 15 )
		VSizer.AddSizer( HSizer, 0, wxALIGN_CENTER_HORIZONTAL | wxBOTTOM, 5 )
		
		SetSizer( VSizer )
		
		Center()
		
		Connect( wxID_VTEXTCTRL_FROTATION,	wxEVT_COMMAND_TEXT_UPDATED,   DispatchFRotationUpdate )
		Connect( wxID_VTEXTCTRL_LPOSITION,	wxEVT_COMMAND_TEXT_UPDATED,   DispatchLPositionUpdate )
		Connect( wxID_VTEXTCTRL_CPOSITION,	wxEVT_COMMAND_TEXT_UPDATED,   DispatchCPositionUpdate )
		Connect( wxID_BUTTON_COLOR,		wxEVT_COMMAND_BUTTON_CLICKED, DispatchColorPick )
		
		If ShowModal() = wxID_OK And Changed Then
			ParentProject.Fractal.Rotate( FractalRotation )
			ParentProject.Camera.LightPosition = LightPosition
			ParentProject.Camera.Locate( CameraPosition )
			ParentProject.Camera.BackgroundColor = BackgroundColor
			ParentProject.MarkChanged()
			
			Destroy()
			
			TMainWindow( GetParent() ).RenderFractal()
			
			Return
		EndIf
		
		Destroy()
	End Method
	
	Function DispatchColorPick( Event:wxEvent )
		Local ImgButton:wxBitmapButton = wxBitmapButton( Event.GetEventObject() )
		
		Local ColorPicker:wxColourDialog = New wxColourDialog.Create( wxWindow( Event.Parent ) )
		
		If ColorPicker.ShowModal() = wxID_OK Then
			Local Color:wxColour = ColorPicker.GetColourData().GetColour()
			
			Local ColorImage:wxImage = wxImage.CreateEmptyImage( 60, 12 )
			ColorImage.SetRGBRange( 0, 0, 60, 12, Color.Red(), Color.Green(), Color.Blue() )
			
			ImgButton.SetBitmapLabel( New wxBitmap.CreateFromImage( ColorImage ) )
			
			TSceneDialog( Event.Parent ).BackgroundColor = Vec3( Color.Red(), Color.Green(), Color.Blue() )
			TSceneDialog( Event.Parent ).Changed = True
		EndIf
	End Function
	
	Function DispatchFRotationUpdate( Event:wxEvent )
		DispatchVectorUpdate( Event, TSceneDialog( Event.Parent ).FractalRotation )
		TSceneDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchLPositionUpdate( Event:wxEvent )
		DispatchVectorUpdate( Event, TSceneDialog( Event.Parent ).LightPosition )
		TSceneDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchCPositionUpdate( Event:wxEvent )
		DispatchVectorUpdate( Event, TSceneDialog( Event.Parent ).CameraPosition )
		TSceneDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchVectorUpdate( Event:wxEvent, Destination:TVector Var )
		Local VectorCtrl:TVectorTextCtrl = TVectorTextCtrl( Event.GetEventObject() )
		Local Values:Float[] = VectorCtrl.GetScalars()
		
		Destination = Vec3( Values[ 0 ], Values[ 1 ], Values[ 2 ] )
	End Function
End Type

Type TRenderDialog Extends wxDialog
	Const wxID_TEXTCTRL_WIDTH:Int	= 300
	Const wxID_TEXTCTRL_HEIGHT:Int	= 301
	Const wxID_CHKBX_SHADOWS:Int		= 302
	Const wxID_CHKBX_ANTIALIASING:Int	= 303
	Const wxID_SPINNER_ITERATIONS:Int	= 304
	Const wxID_SPINNER_EPSILON:Int	= 305
	
	Field ParentProject:TProject
	
	Field ResolutionX:Int
	Field ResolutionY:Int
	Field RenderShadows:Int
	Field AntiAliasing:Int
	Field Iterations:Int
	Field Epsilon:Float
	
	Field Changed:Int
	
	Method OnInit()
		ParentProject = TMainWindow( GetParent() ).CurrentProject
		
		Local VSizer:wxBoxSizer = New wxBoxSizer.Create( wxVERTICAL )
		
		Local HSizer1:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		Local HSizer2:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		Local HSizer3:wxBoxSizer = New wxBoxSizer.Create( wxHORIZONTAL )
		
		Local GSizer:wxFlexGridSizer = New wxFlexGridSizer.CreateRC( 2, 4, 5, 5 )
		
		GSizer.AddGrowableCol( 1 )
		GSizer.AddGrowableCol( 3 )
		
		'--------------- Resolution
		Local Label1:wxStaticText = New wxStaticText.Create( Self, -1, "Image width:" )
		Local Input1:wxTextCtrl   = New wxTextCtrl.Create( Self, wxID_TEXTCTRL_WIDTH, ParentProject.OutputWidth,,, 60 )
		
		Local Label2:wxStaticText = New wxStaticText.Create( Self, -1, "Image height:" )
		Local Input2:wxTextCtrl   = New wxTextCtrl.Create( Self, wxID_TEXTCTRL_HEIGHT, ParentProject.OutputHeight,,, 60 )
		
		GSizer.Add( Label1, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.Add( Input1, 1, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT | wxEXPAND )
		GSizer.Add( Label2, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.Add( Input2, 1, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT | wxEXPAND )
		
		ResolutionX = ParentProject.OutputWidth
		ResolutionY = ParentProject.OutputHeight
		
		'---------------- Special settings
		Iterations		= ParentProject.Fractal.Iterations
		Epsilon		= ParentProject.Fractal.Epsilon
		
		Local Label3:wxStaticText = New wxStaticText.Create( Self, -1, "Iterations:" )
		Local Spinner1:wxSpinCtrl = New wxSpinCtrl.Create( Self, wxID_SPINNER_ITERATIONS, Iterations,,, 70,,, 1, 1000, Iterations )
		Local Label4:wxStaticText = New wxStaticText.Create( Self, -1, "Epsilon:" )
		Local Spinner2:wxSpinCtrl = New wxSpinCtrl.Create( Self, wxID_SPINNER_EPSILON, Int( Ceil( Epsilon*10000 ) ),,, 70,,, 1, 1000, Epsilon*10000 )
		
		GSizer.Add( Label3, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.Add( Spinner1, 1, wxALIGN_CENTER_VERTICAL, 10 )
		GSizer.Add( Label4, 0, wxALIGN_CENTER_VERTICAL | wxALIGN_RIGHT )
		GSizer.Add( Spinner2, 1, wxALIGN_CENTER_VERTICAL, 10 )
		
		HSizer1.AddCustomSpacer( 15, 10 )
		HSizer1.AddSizer( GSizer )
		HSizer1.AddCustomSpacer( 15, 10 )
		
		'--------------- Quality settings
		Local Checkbox1:wxCheckBox = New wxCheckBox.Create( Self, wxID_CHKBX_SHADOWS, "Shadows" )
		Local Checkbox2:wxCheckBox = New wxCheckBox.Create( Self, wxID_CHKBX_ANTIALIASING, "Antialiasing" )
		
		Checkbox1.SetValue( ParentProject.Camera.RenderShadows )
		Checkbox2.SetValue( ParentProject.Camera.AntiAliasing )
		
		RenderShadows	= ParentProject.Camera.RenderShadows
		AntiAliasing	= ParentProject.Camera.AntiAliasing
		
		HSizer2.Add( Checkbox1, 0, wxALIGN_LEFT  | wxLEFT,  30 )
		HSizer2.AddCustomSpacer( 100, 10 )
		HSizer2.Add( Checkbox2, 0, wxALIGN_RIGHT | wxRIGHT, 10 )
		
		'--------------- Buttons
		Local Button1:wxButton = New wxButton.Create( Self, wxID_OK, "OK" )
		Local Button2:wxButton = New wxButton.Create( Self, wxID_CANCEL, "Cancel" )
		
		HSizer3.Add( Button1, 1, wxALIGN_RIGHT | wxRIGHT, 10 )
		HSizer3.Add( Button2, 1, wxALIGN_RIGHT, 10 )
		
		'--------------- Vertical sizer
		VSizer.AddSpacer( 15 )
		VSizer.AddSizer( HSizer1, 1, wxEXPAND )
		VSizer.AddSizer( HSizer2, 0, wxBOTTOM, 10 )
		VSizer.AddSizer( HSizer3, 0, wxALIGN_CENTER_HORIZONTAL | wxBOTTOM, 5 )
		
		SetSizer( VSizer )
		
		Center()
		
		'--------------- Events
		Connect( wxID_TEXTCTRL_WIDTH,  wxEVT_COMMAND_TEXT_UPDATED, DispatchWidthUpdate )
		Connect( wxID_TEXTCTRL_HEIGHT, wxEVT_COMMAND_TEXT_UPDATED, DispatchHeightUpdate )
		
		Connect( wxID_CHKBX_SHADOWS,		wxEVT_COMMAND_CHECKBOX_CLICKED, DispatchShadowClick )
		Connect( wxID_CHKBX_ANTIALIASING,	wxEVT_COMMAND_CHECKBOX_CLICKED, DispatchAntiAliasingClick )
		
		Connect( wxID_SPINNER_EPSILON,	wxEVT_COMMAND_SPINCTRL_UPDATED, DispatchEpsilonSpinner )
		Connect( wxID_SPINNER_ITERATIONS,	wxEVT_COMMAND_SPINCTRL_UPDATED, DispatchIterationSpinner )
		
		If ShowModal() = wxID_OK And Changed Then
			ParentProject.OutputWidth		= ResolutionX
			ParentProject.OutputHeight		= ResolutionY
			ParentProject.Camera.RenderShadows	= RenderShadows
			ParentProject.Camera.AntiAliasing	= AntiAliasing
			ParentProject.Fractal.Iterations	= Iterations
			ParentProject.Fractal.Epsilon	= Epsilon
			ParentProject.MarkChanged()
			
			Destroy()
			
			TMainWindow( GetParent() ).RenderFractal()
			
			Return
		EndIf
		
		Destroy()
	End Method
	
	Function DispatchWidthUpdate( Event:wxEvent )
		Local Dialog:TRenderDialog = TRenderDialog( Event.Parent )
		
		DispatchTextUpdate( Event, Dialog.ResolutionX )
	End Function
	
	Function DispatchHeightUpdate( Event:wxEvent )
		Local Dialog:TRenderDialog = TRenderDialog( Event.Parent )
		
		DispatchTextUpdate( Event, Dialog.ResolutionY )
	End Function
	
	Function DispatchTextUpdate( Event:wxEvent, Variable:Int Var )
		Local TextControl:wxTextCtrl = wxTextCtrl( Event.GetEventObject() )
		Local NewText:String = TextControl.GetLineText( 0 )
		
		If String.FromInt( NewText.ToInt() ) = NewText Or NewText = "" Then
			Variable = NewText.ToInt()
			TRenderDialog( Event.Parent ).Changed = True
		Else
			Local CursorPos:Int = TextControl.GetInsertionPoint()
			TextControl.ChangeValue( Variable )
			TextControl.SetInsertionPoint( CursorPos )
		EndIf
	End Function
	
	Function DispatchShadowClick( Event:wxEvent )
		Local Dialog:TRenderDialog = TRenderDialog( Event.Parent )
		
		Dialog.RenderShadows = Not Dialog.RenderShadows
		TRenderDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchAntiAliasingClick( Event:wxEvent )
		Local Dialog:TRenderDialog = TRenderDialog( Event.Parent )
		
		Dialog.AntiAliasing = Not Dialog.AntiAliasing
		Dialog.Changed = True
	End Function
	
	Function DispatchEpsilonSpinner( Event:wxEvent )
		Local Spinner:wxSpinCtrl = wxSpinCtrl( Event.GetEventObject() )
		
		TRenderDialog( Event.Parent ).Epsilon = Spinner.GetValue()*0.0001
		TRenderDialog( Event.Parent ).Changed = True
	End Function
	
	Function DispatchIterationSpinner( Event:wxEvent )
		Local Spinner:wxSpinCtrl = wxSpinCtrl( Event.GetEventObject() )
		
		TRenderDialog( Event.Parent ).Iterations = Spinner.GetValue()
		TRenderDialog( Event.Parent ).Changed = True
	End Function
	
	Function FloatToString:String( Value:Float, DecimalPlaces:Int )
		Local RawString:String = String.FromFloat( Value )
		
		Return RawString[ .. RawString.Find( "." ) + DecimalPlaces + 1 ]
	End Function
End Type

Type TVectorTextCtrl Extends wxTextCtrl
	Field ScalarCount:Int
	
	Field ValidValue:String
	Field Scalars:Float[]
	Field ScalarString:String[]
	
	Method OnInit()
		Connect( GetId(), wxEVT_COMMAND_TEXT_UPDATED, DispatchTextUpdate )
	End Method
	
	Method SetScalarCount( Count:Int )
		ScalarCount = Count
		
		Scalars		= Scalars[ .. Count ]
		ScalarString	= ScalarString[ .. Count ]
		
		ChangeValue( BuildCaption() )
	End Method
	
	Method SetScalar( Value:Float, Index:Int = 0 )
		Scalars[ Index ]		= Value
		ScalarString[ Index ]	= String.FromFloat( Value )
		
		ChangeValue( BuildCaption() )
	End Method
	
	Method SetScalars( Values:Float[], DecimalPlaces:Int = 7 )
		For Local I:Int = 0 To Min( Values.Length, ScalarCount ) - 1
			Scalars[ I ]		= Values[ I ]
			ScalarString[ I ]		= String.FromFloat( Values[ I ] )
			ScalarString[ I ]		= ScalarString[ I ][ .. ScalarString[ I ].Find( "." ) + DecimalPlaces + 1 ]
		Next
		
		ChangeValue( BuildCaption() )
	End Method
	
	Method GetScalars:Float[]()
		Return Scalars
	End Method
	
	Method SetScalarsFromLabel()
		Local Caption:String = GetLineText( 0 )
		
		If IsValidInput( Caption ) Then
			Caption = Caption[ 1 .. Caption.Length - 1 ]
			
			Local SplitStrings:String[] = Caption.Split( "," )
			
			For Local I:Int = 0 To Self.ScalarCount - 1
				Scalars[ I ]		= SplitStrings[ I ].ToFloat()
				ScalarString[ I ]		= SplitStrings[ I ]
			Next
		EndIf
	End Method
	
	Method BuildCaption:String()
		Local OutString:String
		
		For Local Value:String = EachIn ScalarString
			OutString = OutString + "," + Value
		Next
		
		Return "<" + OutString[ 1 .. ] + ">"
	End Method
	
	Method IsValidInput:Int( Label:String )
		If Chr( Label[ 0 ] ) = "<" And Chr( Label[ Label.Length - 1 ] ) = ">" Then
			Label = Label[ 1 .. Label.Length - 1 ]
			
			Local SplitStrings:String[] = Label.Split( "," )
			
			If SplitStrings.Length = ScalarCount Then
				For Local CurrString:String = EachIn SplitStrings
					If Not IsValidFloat( CurrString ) Return False
				Next
				
				Return True
			EndIf
		EndIf
		
		Return False
	End Method
	
	Function IsValidFloat:Int( Source:String )
		Source = Source.Replace( "-", "" )
		
		If Source.Find( "." ) >= 0 Then
			Local SplitStrings:String[] = Source.Split( "." )
			
			If SplitStrings.Length = 2 Then Return IsValidInteger( SplitStrings[ 0 ] ) And IsValidInteger( SplitStrings[ 1 ] )
		Else
			Return IsValidInteger( Source )
		EndIf
		
		Return False
	End Function
	
	Function IsValidInteger:Int( Source:String )
		For Local I:Int = 0 To Source.Length - 1
			If Source[ I ] < Asc( "0" ) Or Source[ I ] > Asc( "9" ) Then Return False
		Next
		
		Return True
	End Function
	
	Function DispatchTextUpdate( Event:wxEvent )
		Local VectorCtrl:TVectorTextCtrl = TVectorTextCtrl( Event.GetEventObject() )
		
		Local CursorPos:Int = VectorCtrl.GetInsertionPoint()
		
		If Not VectorCtrl.IsValidInput( VectorCtrl.GetLineText( 0 ) ) Then
			VectorCtrl.ChangeValue( VectorCtrl.BuildCaption() )
			VectorCtrl.SetInsertionPoint( CursorPos - 1 )
		Else
			VectorCtrl.SetScalarsFromLabel()
			VectorCtrl.SetInsertionPoint( CursorPos )
			
			Event.Skip()
		EndIf
	End Function
End Type

Type TProject
	Field Name:String
	Field Path:String
	
	Field OutputWidth:Int
	Field OutputHeight:Int
	
	Field Camera:TCamera
	Field Fractal:TJuliaSet
	
	Field Image:TPixmap
	Field Bitmap:wxBitmap
	
	Field Changed:Int
	
	Method SetName( NewName:String )
		Name = NewName
	End Method
	
	Method SetPath( NewPath:String )
		Path = NewPath
	End Method
	
	Method SetStandardValues()
		OutputWidth = 320
		OutputHeight = 240
		
		Camera.SetLightPosition( Vec3( 5, 3, -5 ) )
		Camera.Locate( Vec3( 0, 0, -8 ) )
		Camera.SetBackgroundColor( Vec3( 0, 0, 0 ) )
		Camera.SetShadowFlag( False )
		Camera.SetAntiAliasing( False )
		
		Fractal.Turn( Vec3( 30, 40, 0 ) )
		Fractal.SetMaterial( Vec3( 154, 204, 255 ), 10, 0.45 )
		Fractal.SetApproximation( 8, 0.003 )
		Fractal.SetC( Vec4( 0, 0, 0, 0 ) )
	End Method
	
	Method GetName:String()
		Return Name
	End Method
	
	Method GetImage:TPixmap()
		Return Image
	End Method
	
	Method GetBitmap:wxBitmap()
		Return Bitmap
	End Method
	
	Method MarkChanged()
		Changed = True
		
		MainApp.MainWin.ProjectListControl.SetItemText( MainApp.MainWin.ProjectListControl.FindItem( -1, Name ), Name + "*" )
	End Method
	
	Method MarkUnchanged()
		If Changed Then
			Changed = False
			
			MainApp.MainWin.ProjectListControl.SetItemText( MainApp.MainWin.ProjectListControl.FindItem( -1, Name + "*" ), Name )
		EndIf
	End Method
	
	Method Render( MaxDelay:Int = 50 )
		Image = CreatePixmap( OutputWidth, OutputHeight, PF_RGBA8888 )
		
		Camera.Render( OutputWidth, OutputHeight, Image, RenderFeedback, MaxDelay )
		
		Bitmap = New wxBitmap.CreateFromPixmap( Image )
	End Method
	
	Method RandomizeFractal()
		Fractal.SetMaterial( Vec3( Rand( 64, 255 ), Rand( 64, 255 ), Rand( 64, 255 ) ) )
		Fractal.SetC( Vec4( FractalRand(), FractalRand(), FractalRand(), FractalRand() ) )
	End Method
	
	Method SaveToFile()
		Local OutStream:TProjectStream
		Try
			OutStream = TProjectStream.OpenProject( Path, False, True )
		Catch Error:Object
			RuntimeError "Could not write file '" + Name + "': " + Error.ToString()
		EndTry
		
		If Image Then SavePixmapPNG( Image, "Renderings/" + Name + ".png" )
		
		OutStream.WriteInt( OutputWidth )
		OutStream.WriteInt( OutputHeight )
		OutStream.WriteCamera( Camera )
		OutStream.WriteFractal( Fractal )
		
		OutStream.Close()
	End Method
	
	Method LoadFromFile()
		Local InStream:TProjectStream
		Try
			InStream = TProjectStream.OpenProject( Path, True, False )
		Catch Error:Object
			RuntimeError "Could not read file '" + Name + "': " + Error.ToString()
		EndTry
		
		If FileType( "Renderings/" + Name + ".png" ) Then
			Image = LoadPixmapPNG( "Renderings/" + Name + ".png" )
			Bitmap = New wxBitmap.CreateFromPixmap( Image )
		EndIf
		
		OutputWidth = InStream.ReadInt()
		OutputHeight = InStream.ReadInt()
		Camera = InStream.ReadCamera()
		Fractal = InStream.ReadFractal()
		
		Camera.SetJuliaSet( Fractal )
		
		InStream.Close()
	End Method
	
	Function Create:TProject( Name:String = "Untitled" )
		Local Project:TProject = New TProject
			Project.Name = Name
			Project.Fractal = TJuliaSet.Create()
			Project.Camera = TCamera.Create()
		
		Project.Camera.SetJuliaSet( Project.Fractal )
		
		Return Project
	End Function
	
	Function RenderFeedback:Int( Percentage:Int )
		MainApp.MainWin.ProgressWindow.Gauge.SetValue( Percentage )
		MainApp.Yield()
		
		Return MainApp.MainWin.ProgressWindow.ContinueRender
	End Function
	
	Function FractalRand:Float()
		Return Floor( Rnd( -1, 1 )*100 )/100.0
	End Function
End Type

Type TProjectStream
	Field Stream:TStream
	
	Method WriteCamera( Camera:TCamera )
		WriteFloat( Camera.AOV )
		WriteByte( Camera.RenderShadows )
		WriteByte( Camera.AntiAliasing )
		
		WriteVector( Camera.BackgroundColor )
		WriteVector( Camera.LightPosition )
		
		WriteEntity( Camera )
	End Method
	
	Method WriteFractal( Fractal:TJuliaSet )
		WriteVector( Fractal.C )
		
		WriteInt( Fractal.Iterations )
		WriteFloat( Fractal.Epsilon )
		
		WriteVector( Fractal.DiffuseColor )
		WriteFloat( Fractal.ShadingExponent )
		WriteFloat( Fractal.SpecularHighlight )
		
		WriteEntity( Fractal )
	End Method
	
	Method WriteEntity( Entity:TEntity )
		WriteVector( Entity.Position )
		WriteVector( Entity.Scaling )
		WriteMatrix( Entity.Rotation )
		WriteMatrix( Entity.InvRotation )
	End Method
	
	Method WriteVector( Vector:TVector )
		WriteFloat( Vector.X )
		WriteFloat( Vector.Y )
		WriteFloat( Vector.Z )
		WriteFloat( Vector.W )
	End Method
	
	Method WriteMatrix( Matrix:TMatrix )
		For Local Row:Int = 0 To 3
			For Local Column:Int = 0 To 3
				WriteFloat Matrix.A[ Row, Column ]
			Next
		Next
	End Method
	
	Method WriteInt( Value:Int )
		Stream.WriteInt( Value )
	End Method
	
	Method WriteFloat( Value:Float )
		Stream.WriteFloat( Value )
	End Method
	
	Method WriteByte( Value:Byte )
		Stream.WriteByte( Value )
	End Method
	
	Method ReadCamera:TCamera()
		Local Camera:TCamera = New TCamera
		
		Camera.AOV = Stream.ReadFloat()
		Camera.RenderShadows = ReadByte()
		Camera.AntiAliasing = ReadByte()
		
		Camera.BackgroundColor = ReadVector()
		Camera.LightPosition = ReadVector()
		
		ReadEntity( Camera )
		
		Return Camera
	End Method
	
	Method ReadFractal:TJuliaSet()
		Local Fractal:TJuliaSet = New TJuliaSet
		
		Fractal.C = ReadVector()
		
		Fractal.Iterations = ReadInt()
		Fractal.Epsilon = ReadFloat()
		
		Fractal.DiffuseColor = ReadVector()
		Fractal.ShadingExponent = ReadFloat()
		Fractal.SpecularHighlight = ReadFloat()
		
		ReadEntity( Fractal )
		
		Return Fractal
	End Method
	
	'This method works a bit different - it doesn't create a new entity rather than setting the values of an existing one
	Method ReadEntity( Entity:TEntity )
		Entity.Position = ReadVector()
		Entity.Scaling = ReadVector()
		Entity.Rotation = ReadMatrix()
		Entity.InvRotation = ReadMatrix()
	End Method
	
	Method ReadVector:TVector()
		Local Vector:TVector = New TVector
		
		Vector.X = ReadFloat()
		Vector.Y = ReadFloat()
		Vector.Z = ReadFloat()
		Vector.W = ReadFloat()
		
		Return Vector
	End Method
	
	Method ReadMatrix:TMatrix()
		Local Matrix:TMatrix = New TMatrix
		
		For Local Row:Int = 0 To 3
			For Local Column:Int = 0 To 3
				Matrix.A[ Row, Column ] = ReadFloat()
			Next
		Next
		
		Return Matrix
	End Method
	
	Method ReadInt:Int()
		Return Stream.ReadInt()
	End Method
	
	Method ReadFloat:Float()
		Return Stream.ReadFloat()
	End Method
	
	Method ReadByte:Byte()
		Return Stream.ReadByte()
	End Method
	
	Method Close()
		Stream.Close()
	End Method
	
	Function OpenProject:TProjectStream( Url:Object, Readable:Int = True, Writeable:Int = True )
		Local Stream:TStream = OpenFile( Url, Readable, Writeable )
		
		If Stream Then
			Local ProjectStream:TProjectStream = New TProjectStream
				ProjectStream.Stream = Stream
			
			Return ProjectStream
		EndIf
	End Function
End Type
