Rem
	bbdoc:	Mathematikmodul
	about:	Stell allerlei erweiterte Mathematikfunktionen zu Verfgung
End Rem


Module chaos.mathext
ModuleInfo "Version: 1.00"
ModuleInfo "Author: Noobody"
ModuleInfo "License: Public Domain"
ModuleInfo "Contact: benedikt.bitterli@marbit.ch"

'--------------------------------------------------------------------------------------------------------------------
SuperStrict

Import brl.math

Function Vec2:TVector( X:Float, Y:Float )
	Return TVector.Vector( X, Y, 0 )
End Function

Function Vec3:TVector( X:Float, Y:Float, Z:Float )
	Return TVector.Vector( X, Y, Z )
End Function

Function Vec4:TVector( X:Float, Y:Float, Z:Float, W:Float )
	Return TVector.Vector( X, Y, Z, W )
End Function

Function Mat4:TMatrix( Elements:Int[] )
	Local Matrix:TMatrix = New TMatrix
	
	For Local Row:Int = 0 To 3
		For Local Column:Int = 0 To 3
			Matrix.A[ Row, Column ] = Elements[ Row*4 + Column ]
		Next
	Next
	
	Return Matrix
End Function

Function TFormPoint:TVector( Point:TVector, Source:TEntity = Null, Destination:TEntity = Null )
	Local Result:TVector = Point
	
	If Source <> Null Then
		Source.InvTransform()
		
		Result = Source.InvFinalTransform.VectorMultiply( Result )
	EndIf
	
	If Destination <> Null Then
		Destination.Transform()
		
		Result = Destination.FinalTransform.VectorMultiply( Result )
	EndIf
	
	Return Result
End Function

Function TFormVector:TVector( Vector:TVector, Source:TEntity = Null, Destination:TEntity = Null )
	Local Result:TVector = Vector
	
	If Source <> Null Then
		Source.InvTransform()
		
		Local TFormMat:TMatrix = Source.InvFinalTransform.Copy()
		TFormMat.A[ 0, 3 ] = 0; TFormMat.A[ 1, 3 ] = 0; TFormMat.A[ 2, 3 ] = 0
		
		Result = TFormMat.VectorMultiply( Result )
	EndIf
	
	If Destination <> Null Then
		Destination.Transform()
		
		Local TFormMat:TMatrix = Destination.FinalTransform.Copy()
		TFormMat.A[ 0, 3 ] = 0; TFormMat.A[ 1, 3 ] = 0; TFormMat.A[ 2, 3 ] = 0
		
		Result = TFormMat.VectorMultiply( Result )
	EndIf
	
	Return Result
End Function

Function TFormNormal:TVector( Vector:TVector, Source:TEntity = Null, Destination:TEntity = Null )
	Local Result:TVector = Vector
	
	If Source <> Null Then
		Source.InvTransform()
		
		Local TFormMat:TMatrix = Source.InvFinalTransform.Copy()
		TFormMat.A[ 0, 3 ] = 0; TFormMat.A[ 1, 3 ] = 0; TFormMat.A[ 2, 3 ] = 0
		
		Result = TFormMat.VectorMultiply( Result )
	EndIf
	
	If Destination <> Null Then
		Destination.Transform()
		
		Local TFormMat:TMatrix = Destination.FinalTransform.Copy()
		TFormMat.A[ 0, 3 ] = 0; TFormMat.A[ 1, 3 ] = 0; TFormMat.A[ 2, 3 ] = 0
		
		Result = TFormMat.VectorMultiply( Result )
	EndIf
	
	Result.Normalize()
	
	Return Result
End Function

Type TVector
	Field X:Float
	Field Y:Float
	Field Z:Float
	Field W:Float
	
	Method Add:TVector( B:TVector )
		X :+ B.X
		Y :+ B.Y
		Z :+ B.Z
		
		Return Self
	End Method
	
	Method Substract:TVector( B:TVector )
		X :- B.X
		Y :- B.Y
		Z :- B.Z
		
		Return Self
	End Method
	
	Method Multiply:TVector( Scalar:Float )
		X :* Scalar
		Y :* Scalar
		Z :* Scalar
		
		Return Self
	End Method
	
	Method Divide:TVector( Scalar:Float )
		X :/ Scalar
		Y :/ Scalar
		Z :/ Scalar
		
		Return Self
	End Method
	
	Method DotP:Float( B:TVector )
		Return X*B.X + Y*B.Y + Z*B.Z
	End Method
	
	Method Length:Float()
		Return Sqr( X*X + Y*Y + Z*Z )
	End Method
	
	Method Add4:TVector( B:TVector )
		X :+ B.X
		Y :+ B.Y
		Z :+ B.Z
		W :+ B.W
		
		Return Self
	End Method
	
	Method Substract4:TVector( B:TVector )
		X :- B.X
		Y :- B.Y
		Z :- B.Z
		W :- B.W
		
		Return Self
	End Method
	
	Method Multiply4:TVector( Scalar:Float )
		X :* Scalar
		Y :* Scalar
		Z :* Scalar
		W :* Scalar
		
		Return Self
	End Method
	
	Method Divide4:TVector( Scalar:Float )
		X :/ Scalar
		Y :/ Scalar
		Z :/ Scalar
		W :/ Scalar
		
		Return Self
	End Method
	
	Method DotP4:Float( B:TVector )
		Return X*B.X + Y*B.Y + Z*B.Z + W*B.W
	End Method
	
	Method Length4:Float()
		Return Sqr( X*X + Y*Y + Z*Z + W*W )
	End Method
	
	Method CrossProduct:TVector( B:TVector )
		Local NewX:Float = Y*B.Z - Z*B.Y
		Local NewY:Float = Z*B.X - X*B.Z
		Local NewZ:Float = X*B.Y - Y*B.X
		
		X = NewX
		Y = NewY
		Z = NewZ
		
		Return Self
	End Method
	
	Method Invert:TVector()
		X = 1/X
		Y = 1/Y
		Z = 1/Z
		
		Return Self
	End Method
	
	Method Normalize:TVector()
		Local Length:Float = Sqr( X*X + Y*Y + Z*Z )
		
		X = X/Length
		Y = Y/Length
		Z = Z/Length
		
		Return Self
	End Method
	
	Method Absolute:TVector()
		X = Abs( X )
		Y = Abs( Y )
		Z = Abs( Z )
		
		Return Self
	End Method
	
	Method ToRGB:Int()
		Return ( Int( X ) Shl 16 ) + ( Int( Y ) Shl 8 ) + Int( Z )
	End Method
	
	Method ToRGBA:Int()
		Return ( Int( X ) Shl 24 ) + ( Int( Y ) Shl 16 ) + ( Int( Z ) Shl 8 ) + Int( W )
	End Method
	
	Method Copy:TVector()
		Return TVector.Vector( X, Y, Z, W )
	End Method
	
	Method QuaternionMultiply:TVector( B:TVector )
		Local NewX:Float = X*B.X - Y*B.Y - Z*B.Z - W*B.W
		Local NewY:Float = X*B.Y + B.X*Y + Z*B.W - W*B.Z
		Local NewZ:Float = X*B.Z + B.X*Z + W*B.Y - Y*B.W
		Local NewW:Float = X*B.W + B.X*W + Y*B.Z - Z*B.Y
		
		X = NewX
		Y = NewY
		Z = NewZ
		W = NewW
		
		Return Self
	End Method
	
	Method QuaternionSquare:TVector()
		Local NewX:Float = X*X - Y*Y - Z*Z - W*W
		Local NewY:Float = 2.0*X*Y
		Local NewZ:Float = 2.0*X*Z
		Local NewW:Float = 2.0*X*W
		
		X = NewX
		Y = NewY
		Z = NewZ
		W = NewW
		
		Return Self
	End Method
	
	Function Vector:TVector( X:Float, Y:Float, Z:Float, W:Float = 1 )
		Local Vector:TVector = New TVector
			Vector.X = X
			Vector.Y = Y
			Vector.Z = Z
			Vector.W = W
		
		Return Vector
	End Function
End Type

Type TMatrix
	Field A:Float[ 4, 4 ]
	
	Method Copy:TMatrix()
		Local Matrix:TMatrix = New TMatrix
		
		For Local Row:Int = 0 To 3
			For Local Column:Int = 0 To 3
				Matrix.A[ Row, Column ] = A[ Row, Column ]
			Next
		Next
		
		Return Matrix
	End Method
	
	Method MatrixMultiply:TMatrix( B:TMatrix )
		Local NewA:Float[ 4, 4 ]
		
		For Local Row:Int = 0 To 3
			For Local Column:Int = 0 To 3
				NewA[ Row, Column ]  = A[ Row, 0 ]*B.A[ 0, Column ]
				NewA[ Row, Column ] :+ A[ Row, 1 ]*B.A[ 1, Column ]
				NewA[ Row, Column ] :+ A[ Row, 2 ]*B.A[ 2, Column ]
				NewA[ Row, Column ] :+ A[ Row, 3 ]*B.A[ 3, Column ]
			Next
		Next
		
		A = NewA
		
		Return Self
	End Method
	
	Method VectorMultiply:TVector( B:TVector )
		Local X:Float = A[ 0, 0 ]*B.X + A[ 0, 1 ]*B.Y + A[ 0, 2 ]*B.Z + A[ 0, 3 ]*B.W
		Local Y:Float = A[ 1, 0 ]*B.X + A[ 1, 1 ]*B.Y + A[ 1, 2 ]*B.Z + A[ 1, 3 ]*B.W
		Local Z:Float = A[ 2, 0 ]*B.X + A[ 2, 1 ]*B.Y + A[ 2, 2 ]*B.Z + A[ 2, 3 ]*B.W
		Local W:Float = A[ 3, 0 ]*B.X + A[ 3, 1 ]*B.Y + A[ 3, 2 ]*B.Z + A[ 3, 3 ]*B.W
		
		Return TVector.Vector( X, Y, Z, W )
	End Method
	
	Function Identity:TMatrix()
		Local Matrix:TMatrix = New TMatrix
			Matrix.A[ 0, 0 ] = 1
			Matrix.A[ 1, 1 ] = 1
			Matrix.A[ 2, 2 ] = 1
			Matrix.A[ 3, 3 ] = 1
		
		Return Matrix
	End Function
	
	Function Translation:TMatrix( Vector:TVector )
		Local Matrix:TMatrix = TMatrix.Identity()
			Matrix.A[ 0, 3 ] = Vector.X
			Matrix.A[ 1, 3 ] = Vector.Y
			Matrix.A[ 2, 3 ] = Vector.Z
		
		Return Matrix
	End Function
	
	Function Scale:TMatrix( Vector:TVector )
		Local Matrix:TMatrix = New TMatrix
			Matrix.A[ 0, 0 ] = Vector.X
			Matrix.A[ 1, 1 ] = Vector.Y
			Matrix.A[ 2, 2 ] = Vector.Z
			Matrix.A[ 3, 3 ] = 1
		
		Return Matrix
	End Function
	
	Function RotationX:TMatrix( Angle:Float )
		Local Matrix:TMatrix = TMatrix.Identity()
			Matrix.A[ 1, 1 ] =  Cos( Angle )
			Matrix.A[ 1, 2 ] =  Sin( Angle )
			Matrix.A[ 2, 1 ] = -Sin( Angle )
			Matrix.A[ 2, 2 ] =  Cos( Angle )
		
		Return Matrix
	End Function
	
	Function RotationY:TMatrix( Angle:Float )
		Local Matrix:TMatrix = TMatrix.Identity()
			Matrix.A[ 0, 0 ] =  Cos( Angle )
			Matrix.A[ 0, 2 ] =  Sin( Angle )
			Matrix.A[ 2, 0 ] = -Sin( Angle )
			Matrix.A[ 2, 2 ] =  Cos( Angle )
		
		Return Matrix
	End Function
	
	Function RotationZ:TMatrix( Angle:Float )
		Local Matrix:TMatrix = TMatrix.Identity()
			Matrix.A[ 0, 0 ] =  Cos( Angle )
			Matrix.A[ 0, 1 ] =  Sin( Angle )
			Matrix.A[ 1, 0 ] = -Sin( Angle )
			Matrix.A[ 1, 1 ] =  Cos( Angle )
		
		Return Matrix
	End Function
	
	Function RotationXYZ:TMatrix( Rotation:TVector )
		Return TMatrix.RotationZ( Rotation.Z ).MatrixMultiply( RotationY( Rotation.Y ).MatrixMultiply( RotationX( Rotation.X ) ) )
	End Function
	
	Function RotationZYX:TMatrix( Rotation:TVector )
		Return TMatrix.RotationX( Rotation.X ).MatrixMultiply( RotationY( Rotation.Y ).MatrixMultiply( RotationZ( Rotation.Z ) ) )
	End Function
	
	Rem
	Method DebugDraw( X:Int, Y:Int )
		DrawLine X - 5, Y - 6, X - 5, Y + 94 'Hardcoding \o/
		DrawLine X - 5, Y - 6, X + 2, Y - 6
		DrawLine X - 5, Y + 94, X + 2, Y + 94
		
		Local ColumnWidth:Int
		For Local Column:Int = 0 To 3
			For Local Row:Int = 0 To 3
				Local Element:Float = A[ Row, Column ]
				
				If Element < 0 Then
					DrawText Element, X, Y + Row*25
				Else
					DrawText Element, X + TextWidth( "-" ), Y + Row*25
				EndIf
				
				If TextWidth( Element ) > ColumnWidth Then ColumnWidth = TextWidth( Element )
			Next
			
			X :+ 20 + ColumnWidth
			ColumnWidth = 0
		Next
		
		DrawLine X - 5, Y - 6, X - 5, Y + 94
		DrawLine X - 5, Y - 6, X - 12, Y - 6
		DrawLine X - 5, Y + 94, X - 12, Y + 94
	End Method
	End Rem
End Type

Type TEntity
	Field Parent:TEntity
	Field FinalTransform:TMatrix
	Field InvFinalTransform:TMatrix
	
	Field Position:TVector
	Field Scaling:TVector
	Field Rotation:TMatrix
	Field InvRotation:TMatrix
	
	Method New()
		Locate( Vec3( 0, 0, 0 ) )
		Rotate( Vec3( 0, 0, 0 ) )
		Scale ( Vec3( 1, 1, 1 ) )
	End Method
	
	Method Turn( Vector:TVector )
		Rotation = TMatrix.RotationZYX( Vector ).MatrixMultiply( Rotation )
		InvRotation = InvRotation.MatrixMultiply( TMatrix.RotationXYZ( Vector.Multiply( -1 ) ) )
	End Method
	
	Method Move( Vector:TVector )
		Position.Add( Vector )
	End Method
	
	Method Rotate( Vector:TVector, GlobalSystem:Int = False )
		If GlobalSystem And Parent <> Null Then
			Throw "Currently not supported!"
		Else
			Rotation = TMatrix.RotationZYX( Vector )
			InvRotation = TMatrix.RotationXYZ( Vector.Multiply( -1 ) )
		EndIf
	End Method
	
	Method Locate( Vector:TVector, GlobalSystem:Int = False )
		If GlobalSystem And Parent <> Null Then
			Parent.Transform()
			
			Position = Parent.FinalTransform.VectorMultiply( Vector )
		Else
			Position = Vector
		EndIf
	End Method
	
	Method Scale( Vector:TVector )
		Scaling = Vector
	End Method
	
	Method SetParent( Entity:TEntity )
		Parent = Entity
	End Method
	
	Method GetRotation:TVector()
		Local Pitch:Float = ATan2( Rotation.A[ 1, 2 ], Rotation.A[ 2, 2 ] )
		Local Yaw:Float = ASin( -Rotation.A[ 0, 2 ] )
		Local Roll:Float = ATan2( Rotation.A[ 0, 1 ], Rotation.A[ 0, 0 ] )
		
		Return Vec3( Pitch, -Yaw, Roll )
	End Method
	
	Method Transform()
		FinalTransform = TMatrix.Translation( Position.Copy().Multiply( -1 ) ).Copy().MatrixMultiply( Rotation.Copy().MatrixMultiply( TMatrix.Scale( Scaling.Invert() ) ) )
		
		If Parent Then
			Parent.Transform()
			
			FinalTransform.MatrixMultiply( Parent.FinalTransform )
		EndIf
	End Method
	
	Method InvTransform()
		InvFinalTransform = TMatrix.Translation( Position ).MatrixMultiply( InvRotation.Copy().MatrixMultiply( TMatrix.Scale( Scaling ) ) )
		
		If Parent Then
			Parent.InvTransform()
			
			InvFinalTransform = Parent.InvFinalTransform.Copy().MatrixMultiply( InvFinalTransform )
		EndIf
	End Method
End Type

Type THermiteSpline
	Global Polynomial:TMatrix = Mat4( ..
		[  2, -3,  0,  1, ..
		  -2,  3,  0,  0, ..
		   1, -2,  1,  0, ..
		   1, -1,  0,  0  ] )
	
	Field NodePosition:TVector[]
	Field NodeTangent:TVector[]
	
	Method AddNode( Position:TVector )
		NodePosition = NodePosition + [ Position ]
		
		If NodePosition.Length > 2 Then
			Local Index:Int		= NodePosition.Length
			Local Node0:TVector	= NodePosition[ 0 ]
			Local Node1:TVector	= NodePosition[ 1 ]
			Local Node2:TVector	= NodePosition[ ( Index*2 - 3 ) Mod Index ]
			Local Node3:TVector	= NodePosition[ Index - 2 ]
			Local Node4:TVector	= NodePosition[ Index - 1 ]
			
			Local T0:TVector = Node1.Copy().Substract4( Node4 ).Divide4( 2.0 )
			Local T2:TVector = Node4.Copy().Substract4( Node2 ).Divide4( 2.0 )
			Local T3:TVector = Node0.Copy().Substract4( Node3 ).Divide4( 2.0 )
			
			NodeTangent  = NodeTangent  + [ T3 ]
			NodeTangent[ 0 ] = T0
			NodeTangent[ Index - 2 ] = T2
		Else
			NodeTangent = NodeTangent + [ Vec4( 0, 0, 0, 0 ) ]
		EndIf
	End Method
	
	Method UpdateChanges()
		Local Boundary:Int = NodePosition.Length
		
		For Local NodeIndex:Int = 0 To Boundary - 1
			Local Prev:TVector = NodePosition[ ( NodeIndex - 1 + Boundary ) Mod Boundary ]
			Local Succ:TVector = NodePosition[ ( NodeIndex + 1 ) Mod Boundary ]
			
			NodeTangent[ NodeIndex ] = Succ.Copy().Substract4( Prev ).Divide4( 2.0 )
		Next
	End Method
	
	Method Interpolate:TVector( T:Float, UseZ:Int = False, UseW:Int = False )
		Local Index1:Int = Floor( T ) Mod NodePosition.Length
		Local Index2:Int = ( Index1 + 1 ) Mod NodePosition.Length
		
		Local P1:TVector = NodePosition[ Index1 ]
		Local T1:TVector = NodeTangent[  Index1 ]
		Local P2:TVector = NodePosition[ Index2 ]
		Local T2:TVector = NodeTangent[  Index2 ]
		
		T = T Mod 1.0
		
		Local S:TVector = Vec3( T*T*T, T*T, T )
		Local CX:TVector = Vec4( P1.X, P2.X, T1.X, T2.X )
		Local CY:TVector = Vec4( P1.Y, P2.Y, T1.Y, T2.Y )
		
		Local ResultX:Float = Polynomial.VectorMultiply( S ).DotP4( CX )
		Local ResultY:Float = Polynomial.VectorMultiply( S ).DotP4( CY )
		Local ResultZ:Float = 0
		Local ResultW:Float = 1
		
		If UseZ Then
			Local CZ:TVector = Vec4( P1.Z, P2.Z, T1.Z, T2.Z )
			ResultZ = Polynomial.VectorMultiply( S ).DotP4( CZ )
			
			If UseW Then
				Local CW:TVector = Vec4( P1.W, P2.W, T1.W, T2.W )
				ResultW = Polynomial.VectorMultiply( S ).DotP4( CW )
			EndIf
		EndIf
		
		Return Vec4( ResultX, ResultY, ResultZ, ResultW )
	End Method
End Type